#include "svnmodel.h"
#include <QGraphicsScene>
#include <QProcess>
#include <QXmlDefaultHandler>
#include <QXmlParseException>
#include <QDateTime>
#include <QEvent>
#include <QMouseEvent>
#include <QUrl>
#include <QDebug>
#include <cmath>
#include <climits>

class SvnXmlHandler : public QXmlDefaultHandler {
public:
    enum State {None, Author, Date, Msg };
    State state;
    SvnEntry entry;
    QString date;
    SvnModel& model;

    SvnXmlHandler(SvnModel& m) :state(None), model(m) {}
    
    bool startElement(const QString&, const QString& localName, const QString&,
            const QXmlAttributes& atts) {
        if (localName == "logentry") {
            entry.author = date = entry.msg = "";
            entry.time = 0;
            entry.revision = atts.value("revision").toUInt();
            state = None;
        } else if (localName == "author") {
            state = Author;
        } else if (localName == "date") {
            state = Date;
        } else if (localName == "msg") {
            state = Msg;
        }
        return true;
    }
    bool endElement(const QString&, const QString& localName, const QString &) {
        if (localName == "logentry") {
            entry.time = QDateTime::fromString(date, Qt::ISODate).toTime_t();
            model.addEntry(entry);
        }
        return true;
    }
    bool characters(const QString& ch) {
        switch (state) {
        case Author: entry.author += ch; break;
        case Date: date += ch; break;
        case Msg: entry.msg += ch; break;
        default: break;
        }
        return true;
    }
    bool endDocument() {
        qDebug() << "ready";
        return true;
    }
    bool error(const QXmlParseException& exception) {
        qDebug() << exception.message();
        return false;
    }
    bool fatalError(const QXmlParseException& exception) {
        qDebug() << exception.message();
        return false;
    }
    bool warning(const QXmlParseException& exception) {
        qDebug() << exception.message();
        return false;
    }
};

class SvnModelPriv {
public:
    QProcess loader;
    QXmlSimpleReader reader;
    SvnXmlHandler handler;
    QXmlInputSource* source;

    SvnModelPriv(SvnModel&m) :handler(m), source(0) {
        reader.setContentHandler(&handler);
        reader.setErrorHandler(&handler);
    }
    ~SvnModelPriv() {
        if (source) delete source;
    }
};
SvnModel::SvnModel(QGraphicsScene* s) :QObject(s), scene(s), mintime(UINT_MAX),
        maxtime(0), zoom(1), minzoom(0.1), maxzoom(10), offset(0) {
    minzoom = maxzoom = 1;
    width = 1;
    hadinput = false;
    p = new SvnModelPriv(*this);
    connect(&p->loader, SIGNAL(readyReadStandardOutput()),
        this, SLOT(parseInputIncrementally()));
    connect(&p->loader, SIGNAL(finished(int, QProcess::ExitStatus)),
        this, SLOT(processFinished(int, QProcess::ExitStatus)));
}
SvnModel::~SvnModel() {
    qDebug() << p->loader.state();
    delete p;
}
void
SvnModel::show(const QString& svn) {
    if (svn == svndir) return;
    svndir = svn;
    p->loader.terminate();
    p->loader.waitForFinished(1000);

    QList<QGraphicsItem *> items = scene->items();
    for (int i=0; i<items.size(); ++i) {
        scene->removeItem(items[i]);
    }
    mintime = UINT_MAX;
    maxtime = 0;
    hadinput = false;
    entries.clear();

    QStringList args;
    args << "log" << "--xml" << "--limit" << "500" <<
        "svn://anonsvn.kde.org/home/kde/" + svndir;
    p->loader.start("svn", args, QIODevice::ReadOnly);
    if (p->source) {
        delete p->source;
        p->source = 0;
    }
}
void
SvnModel::parseInputIncrementally() {
    if (p->source) {
        p->reader.parseContinue();
    } else {
        p->source = new QXmlInputSource(&p->loader);
        p->reader.parse(p->source, true);
    }
}
void
SvnModel::processFinished(int exitCode, QProcess::ExitStatus /*exitStatus*/) {
    qDebug() << "finished " << exitCode;
    offset = -zoom*mintime;
}
#include <QPen>
QColor
getColor(int i) {
    int f=255, h=127, e=0;
    switch(i%12) {
    case 0: return QColor(f, e, e, h);
    case 1: return QColor(e, f, e, h);
    case 2: return QColor(e, e, f, h);
    case 3: return QColor(f, f, e, h);
    case 4: return QColor(f, e, f, h);
    case 5: return QColor(e, f, f, h);
    case 6: return QColor(f, h, e, h);
    case 7: return QColor(f, e, h, h);
    case 8: return QColor(h, f, e, h);
    case 9: return QColor(h, e, f, h);
    case 10: return QColor(e, f, h, h);
    case 11: return QColor(e, h, f, h);
    }
    return QColor(f,f,f,f);
}
void
SvnModel::addEntry(const SvnEntry& entry) {
    if (mintime > entry.time) {
        mintime = entry.time;
    }
    if (maxtime < entry.time) {
        maxtime = entry.time;
    }
    if (!hadinput) {
    zoom = minzoom;
        if (updateView(0,0,0,0)) {
            updateCoords();
        }
    }

    SvnGraphicsItem* item = new SvnGraphicsItem(entry, scene, this);
    item->setAcceptsHoverEvents(true);
    item->setRect(0, 0,10,50);
    item->setVisible(true);
    if (!colors.contains(entry.author)) {
        colors.insert(entry.author, getColor(colors.size()));
    }
    item->setBrush(colors[entry.author]);
    item->setPen(QPen(Qt::black));
    scene->setSceneRect(QRectF(0,0,10,20));
    entries.prepend(item);
}
bool
SvnModel::eventFilter(QObject *obj, QEvent *event) {
    if (event->type() == QEvent::MouseMove) {
        QMouseEvent* e = static_cast<QMouseEvent*>(event);
        if (e->buttons() & Qt::LeftButton) {
            int x = e->x();
            int y = e->y();
            if (updateView(x, lasty, lastx, lasty)) {
                updateCoords();
            }
            lastx = x;
            lasty = y;
        }
        hadinput = true;
    } else if (event->type() == QEvent::MouseButtonPress) {
        QMouseEvent* e = static_cast<QMouseEvent*>(event);
        lastx = e->x();
        lasty = e->y();
        hadinput = true;
    } else if (event->type() == QEvent::Wheel) {
        QWheelEvent* e = static_cast<QWheelEvent*>(event);
        if (updateView(e->x(), 0, e->x(), -e->delta()/5)) {
            updateCoords();
        }
        hadinput = true;
    } else if (event->type() == QEvent::Resize) {
        QResizeEvent* e = static_cast<QResizeEvent*>(event);
        width = e->size().width();
        if (updateView(0,0,0,0)) {
            updateCoords();
        }
    }
    return QObject::eventFilter(obj, event);
}
void
SvnModel::updateLimits() {
    if (width <= 0) width = 1;
    uint dt = maxtime - mintime;
    minzoom = 1.0*(width-50)/dt;
    maxzoom = 1000*minzoom;
}
bool
SvnModel::updateView(int x1, int y1, int x2, int y2) {
    if (mintime >= maxtime) {
        zoom = 1;
        offset = 0;
        return false;
    }
    double newzoom = zoom;
    double v = (x1-offset)/zoom;
    if (v < mintime) v = mintime;
    if (v > maxtime) v = maxtime;
    updateLimits();

    int dy = y1-y2;
    if (dy) {
        newzoom *= exp(dy/100.0);
    }
    if (newzoom < minzoom) newzoom = minzoom;
    if (newzoom > maxzoom) newzoom = maxzoom;

    int margin = 25;
    double minv = mintime + (x1-margin)/newzoom;
    if (v < minv) v = minv;
    double maxv = maxtime - (width-x1-margin)/newzoom;
    if (v > maxv) v = maxv;

    double newoffset = x1+x1-x2 - newzoom*v;
    if (newzoom != zoom || newoffset != offset) {
        zoom = newzoom;
        offset = newoffset;
        return true;
    }
    return false;
}
void
SvnModel::updateCoords() {
    for (int i=0; i<entries.size(); ++i) {
        entries[i]->setPosParams(zoom, offset);
    }
}
void
SvnModel::handleClick(SvnGraphicsItem* item) {
    QString text = "http://websvn.kde.org/" + svndir + "?rev="
        + QString::number(item->entry.revision);
    text = "<table><tr><td>Revision:</td><td>"
        + QString::number(item->entry.revision)
        + "</td></tr><tr><td>Date:</td><td>"
        + QDateTime::fromTime_t(item->entry.time).toString()
        + "</td></tr><tr><td>Author:</td><td>"+item->entry.author
        + "</td></tr><tr><td colspan='2'>"+item->entry.msg+"</td></tr></table>";
    emit itemActivated(text);
}
void
SvnGraphicsItem::setPosParams(double zoom, double offset) {
    setPos(zoom*entry.time+offset, 0);
}
void
SvnGraphicsItem::hoverEnterEvent(QGraphicsSceneHoverEvent* event) {
    model->handleClick(this);
}
